/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package core;
import java.util.ArrayList;
import java.util.List;
import java.lang.reflect.*;

import ui.DTNSimTextUI;
import gui.DTNSimGUI;

/**
 * Simulator's main class 
 */
public class DTNSim {
	/** If this option ({@value}) is given to program, batch mode and
	 * Text UI are used*/
	public static final String BATCH_MODE_FLAG = "-b";
	
	/** Name of the static method that all resettable classes must have
	 * @see #registerForReset(String) */
	public static final String RESET_METHOD_NAME = "reset";
	private static List<Class<?>> resetList = new ArrayList<Class<?>>();
	
	/**
	 * Starts the user interface with given arguments.
	 * If first argument is {@link #BATCH_MODE_FLAG}, the text UI is started.
	 * Next argument is the settings file for the simulation run (if any).
	 * For text UI, also number of runs can be given after the setting file 
	 * name. Different settings from run arrays are used for different runs (see
	 * {@link Settings#setRunIndex(int)}). For GUI mode, the number after 
	 * settings file (if given) is the run index to use for that run.
	 * @param args Command line arguments
	 */
	public static void main(String[] args) {
		boolean batchMode = false;
		String confFile = null;
		int nrofRuns = 1;
		// set US locale to parse decimals in consistent way
		java.util.Locale.setDefault(java.util.Locale.US);
		
		if (args.length > 0) {
			if (args[0].equals(BATCH_MODE_FLAG)) {
				batchMode = true;
				if (args.length > 1) {
					confFile = args[1];
				}
				if (args.length > 2) {
					nrofRuns = parseNrofRuns(args[2]);
				}
			}
			else {
				confFile = args[0];
				if (args.length > 1) { // use specific run index
					int runIndex = parseNrofRuns(args[1]);
					Settings.setRunIndex(runIndex);
				}
			}
		}
		
		initSettings(confFile);
		
		if (batchMode) {
			long startTime = System.currentTimeMillis();
			for (int i=0; i<nrofRuns; i++) {
				if (nrofRuns > 1) {
					print("Run " + (i+1) + "/" + nrofRuns);
				}
				Settings.setRunIndex(i);
				resetForNextRun();
				new DTNSimTextUI().start();
			}
			double duration = (System.currentTimeMillis() - startTime)/1000.0;
			print("---\nAll done in " + String.format("%.2f", duration) + "s");
		}
		else {
			new DTNSimGUI().start();
		}
	}
	
	/**
	 * Initializes Settings
	 * @param confFile Where to read additional settings (in addition to
	 * default settings) or null if there are no additional settings
	 */
	private static void initSettings(String confFile) {
		try {
			Settings.init(confFile);
		}
		catch (SettingsError er) {
			System.err.println("Failed to load settings: " + er);
			System.err.println("Caught at " + er.getStackTrace()[0]);
			System.exit(-1);
		}
	}
	
	/**
	 * Registers a class for resetting. Reset is performed after every
	 * batch run of the simulator to reset the class' state to initial
	 * state. All classes that have static fields that should be resetted
	 * to initial values between the batch runs should register using 
	 * this method. The given class must have a static implementation
	 * for the resetting method (a method called {@value #RESET_METHOD_NAME} 
	 * without any parameters).
	 * @param className Full name (i.e., containing the packet path) 
	 * of the class to register. For example: <code>core.SimClock</code> 
	 */
	public static void registerForReset(String className) {
		Class<?> c = null;
		try {
			c = Class.forName(className);
			c.getMethod(RESET_METHOD_NAME);
		} catch (ClassNotFoundException e) {
			System.err.println("Can't register class " + className + 
					" for resetting; class not found");
			System.exit(-1);
					
		}
		catch (NoSuchMethodException e) {
			System.err.println("Can't register class " + className + 
			" for resetting; class doesn't contain resetting method");
			System.exit(-1);
		}
		resetList.add(c);
	}
	
	/**
	 * Resets all registered classes.
	 */
	private static void resetForNextRun() {
		for (Class<?> c : resetList) {
			try {
				Method m = c.getMethod(RESET_METHOD_NAME);
				m.invoke(null);
			} catch (Exception e) {
				System.err.println("Failed to reset class " + c.getName());
				e.printStackTrace();
				System.exit(-1);
			}
		}
	}
	
	/**
	 * Parses the number of runs from command line argument
	 * @param arg The argument to parse
	 * @return The value as integer
	 */
	private static int parseNrofRuns(String arg) {
		int val = 1;
		
		try {
			val = Integer.parseInt(arg);
		} catch (NumberFormatException e) {
			System.err.println("Invalid numeric argument '" + arg + "' for" +
					" number of runs");
			System.exit(-1);
		}
		
		if (val < 1) {
			val = 1;
		}
		
		return val;
	}
	
	/**
	 * Prints text to stdout
	 * @param txt Text to print
	 */
	private static void print(String txt) {
		System.out.println(txt);
	}
}
